﻿using SFML.Graphics;
using SFML.Window;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace BobbleCrowd
{
    public class ConvexCollider : ACollider
    {
        //____________MEMBER / GETTER / SETTER________________

        private List<Vector2f> vertices;

        public Vector2f this[int index]
        {
            get { return vertices[index]; }
            set { vertices[index] = value;}
        }

        private Matrix transformMatrix;
        private Vector2f[] normals;

        public bool redundantNormals = false;

        public Vector2f GetNormal(int index)
        {
            return normals[index];
        }
        public void SetNormal(int index, Vector2f normal)
        {
            System.Diagnostics.Debug.Assert(MathUtil.NearEqual(Vec2f.Length(normal), 1.0f));

            normals[index] = normal;
        }

        public int VertexCount
        {
            get;
            private set;
        }

        //______________CONSTRUCTOR_________________________ 

        public ConvexCollider(params Vector2f[] vertices)
            : base(ECollidable.Convex)
        {
            //+ 2 for common center and the second Vector2f twice to close the "gap"
            this.vertices = new List<Vector2f>(vertices.Length);

            for (uint i = 0; i < vertices.Length; i++)
            {
                this.vertices.Add(vertices[i]);
                position += vertices[i];
            }
              

            //Compute the center of all vertices:
            position /= vertices.Length;

            transformMatrix = Matrix.Identity;

            normals = new Vector2f[this.vertices.Count];

            this.VertexCount = this.vertices.Count;

            ComputeNormals();
            ComputeRadius();
            CalculateBoundingRectangle();
        }

        public ConvexCollider(bool redundant, params Vector2f[] vertices)
            : this(vertices)
        {
            this.redundantNormals = redundant;
        }



        //______________TRANSFORM METHODS_________________________

        public override void Rotate(float deltaRad)
        {
            Rotate(deltaRad, position);
        }

        public override void Move(float deltaX, float deltaY)
        {
            for (int i = 0; i < vertices.Count; i++)
                this.vertices[i] += new Vector2f(deltaX, deltaY);
            
            position.X += deltaX;
            position.Y += deltaY;

            MoveBoundingRectangle();
        }

        public override void ScaleUp(float scaleFactor)
        {
            ScaleUp(scaleFactor, position);
        }


        public void Rotate(float deltaRad, Vector2f rotaOrigin)
        {

            transformMatrix.ConvertToRotation(deltaRad);

            for (int i = 0; i < vertices.Count; i++)
            {
                vertices[i] = transformMatrix.Transform(vertices[i] - rotaOrigin) + rotaOrigin;

                if (i < vertices.Count)
                    normals[i] = transformMatrix.Transform(normals[i]);
            }

            if (!Vec2f.NearEqual(position, rotaOrigin))
                position = transformMatrix.Transform(position - rotaOrigin) + rotaOrigin;

            rotation += deltaRad;
            direction = Vec2f.DirectionFrom(rotation);

        }

        public void ScaleUp(float scaleFactor, Vector2f scaleOrigin)
        {
            transformMatrix.ConvertToScale(scaleFactor, scaleFactor);

            for (int i = 0; i < vertices.Count; i++)
                vertices[i] = transformMatrix.Transform(vertices[i] - scaleOrigin) + scaleOrigin;

            if (!Vec2f.NearEqual(scaleOrigin, position))
                position = transformMatrix.Transform(position - scaleOrigin) + scaleOrigin;

            radius *= scaleFactor;
        }

        //___________________DRAWING METHODS_______________
        /*
        public override void Draw(RenderTarget target)
        {
            target.Draw(vertices);

            /*
            for (int i = 0; i < VertexCount; i++)
            {
                CircleShape tmpShape = new CircleShape(5.0f);
                tmpShape.Origin = new Vector2f(5.0f, 5.0f);
                tmpShape.Position = this[i];

                target.Draw(tmpShape);
            }
            */

            /*
            CircleShape radiusShape = new CircleShape(radius);
            radiusShape.Origin = new Vector2f(radius, radius);
            radiusShape.Position = position;
            radiusShape.FillColor = new Color(255, 0, 0, 55);

            target.Draw(radiusShape);
            */

            //position:

            /*
            CircleShape posShape = new CircleShape(5.0f);
            posShape.Origin = new Vector2f(5.0f, 5.0f);
            posShape.Position = position;
            posShape.FillColor = new Color(255, 255, 0, 255);

            target.Draw(posShape);
            */

            ///bounding Rectanle:

            /*
            RectangleShape rectShape = new RectangleShape(new Vector2f(boundingRect.Width, boundingRect.Height));
            rectShape.Position = new Vector2f(boundingRect.Left, boundingRect.Top);
            rectShape.FillColor = new Color(255, 255, 255, 125);

            target.Draw(rectShape);
           */
      //  }
    
        /*
        public override void Draw(RenderTarget target, RenderStates states)
        {
            target.Draw(vertices, states);
        }*/

    
        //___________________INTERSECTION METHODS_______________

        public override IntersectionData Intersects(ACollider other)
        {
            if (other.CollidableType == ECollidable.Convex)
                return SatConvex(other as ConvexCollider);

            else if (other.CollidableType == ECollidable.Circle)
                return SatCircle(other as CircleCollider);

           // else if (other.CollidableType == ECollidable.Concave)
          //      return (other as ConcaveCollider).SatConvex(this, true);


            return new IntersectionData();
        }

        //TODO: optimaze, make more readable:
        public IntersectionData SatCircle(CircleCollider other)
        {
            if (this.IntersectsBounding(other))
            {
                float depth = float.PositiveInfinity;
                Vector2f dir = Vec2f.Zero;

                float nearestDistance = float.PositiveInfinity;
                Vector2f nearestPos = Vec2f.Zero;


                int counter = redundantNormals ? vertices.Count / 2 : vertices.Count;
                for (int i = 0; i < vertices.Count; i++)
                {
                    float tmpDist = Vec2f.EuclidianDistance(this[i], other.Position);

                    if (tmpDist < nearestDistance)
                    {
                        nearestDistance = tmpDist;
                        nearestPos = this[i];
                    }

                    //save half the computations when the polygon has normals twice
                    if (!redundantNormals || i < counter)
                    {
                        float aMin = 0;
                        float aMax = 0;

                        findMinMax(ref aMin, ref aMax, ref normals[i]);

                        float bMin = Vec2f.Dot(other.Position - normals[i] * other.Radius, normals[i]);
                        float bMax = Vec2f.Dot(other.Position + normals[i] * other.Radius, normals[i]);

                        if (aMin > bMax || bMin > aMax)
                            return new IntersectionData();

                        float help = aMax == bMax ? Math.Max(aMin - aMax, bMin - bMax) : aMax > bMax ? aMin - bMax : bMin - aMax;

                        if (help < 0)
                            help = -help;

                        if (help < depth)
                        {
                            depth = help;
                            dir = -normals[i];
                        }
                    }
                }

                //projection to the nearest point to the center of the circle:
                Vector2f tmpDir = Vec2f.Normalized(other.Position - nearestPos, nearestDistance);

                float aMin_ = 0;
                float aMax_ = 0;

                findMinMax(ref aMin_, ref aMax_, ref tmpDir);

                float bMin_ = Vec2f.Dot(other.Position - tmpDir * other.Radius, tmpDir);
                float bMax_ = Vec2f.Dot(other.Position + tmpDir * other.Radius, tmpDir);

                if (aMin_ > bMax_ || bMin_ > aMax_)
                    return new IntersectionData();

                float help_ = aMax_ == bMax_ ? Math.Max(aMin_ - aMax_, bMin_ - bMax_) : aMax_ > bMax_ ? aMin_ - bMax_ : bMin_ - aMax_;

                if (help_ < 0)
                    help_ = -help_;

                if (help_ < depth)
                {
                    depth = help_;
                    dir = -tmpDir;
                }

                //direction and depth found now


                if (Vec2f.Dot(dir, other.Position - position) < 0)
                    dir = -dir;

                return new IntersectionData(depth, dir);
            }

            else
                return new IntersectionData();
        }

        public IntersectionData SatConvex(ConvexCollider other)
        {
            if (this.IntersectsBounding(other))
            {

                float peneValue = float.PositiveInfinity;
                Vector2f sepaNormal = new Vector2f(0, 0);


                int counter = redundantNormals ? vertices.Count / 2 : vertices.Count;
                for (int i = 0; i < counter; i++)
                {
                    float aMin = 0;
                    float aMax = 0;
                    findMinMax(ref aMin, ref aMax, ref normals[i]);

                    float bMin = 0;
                    float bMax = 0;
                    other.findMinMax(ref bMin, ref bMax, ref normals[i]);

                    if (aMin > bMax || bMin > aMax)
                        return new IntersectionData();

                    float help = aMax == bMax ? Math.Max(aMin - aMax, bMin - bMax) : aMax > bMax ? aMin - bMax : bMin - aMax;

                    if (help < 0)
                        help = -help;

                    if (help < peneValue)
                    {
                        peneValue = help;
                        sepaNormal = -normals[i];
                    }

                }

                //same as above for the normals of the second polygon
                int counter2 = other.redundantNormals ? other.vertices.Count / 2 : other.vertices.Count;
                for (int i = 0; i < counter2; i++)
                {
                    float aMin = 0;
                    float aMax = 0;

                    float bMin = 0;
                    float bMax = 0;

                    findMinMax(ref aMin, ref aMax, ref other.normals[i]);
                    other.findMinMax(ref bMin, ref bMax, ref other.normals[i]);

                    if (aMin > bMax || bMin > aMax)
                        return new IntersectionData();

                    float help = aMax == bMax ? Math.Max(aMin - aMax, bMin - bMax) : aMax > bMax ? aMin - bMax : bMin - aMax;

                    if (help < 0)
                        help = -help;

                    if (help < peneValue)
                    {
                        peneValue = help;
                        sepaNormal = -other.normals[i];
                    }
                }

                if (Vec2f.Dot(sepaNormal, other.position - position) < 0)
                    sepaNormal = -sepaNormal;

                return new IntersectionData(peneValue, sepaNormal);
            }

            else
                return new IntersectionData();
        }


        //__________________PRIVATE METHODS_________________________

        private void findMinMax(ref float min, ref float max, ref Vector2f normal)
        {
            min = float.PositiveInfinity;
            max = float.NegativeInfinity;

            for (int i = 0; i < this.vertices.Count; i++)
            {
                float current = Vec2f.Dot(this[i], normal);

                if (current < min)
                    min = current;

                if (current > max)
                    max = current;
            }
        }

        private void ComputeNormals()
        {
            for (int i = 0; i < this.vertices.Count; i++)
            {
                normals[i] = Vec2f.RightNormal(this[(i + 1) % this.vertices.Count] - this[i]);
            }
        }

        private void ComputeRadius()
        {
            float tmpRadius = float.NegativeInfinity;

            for (int i = 0; i < this.vertices.Count; i++)
            {
                float tmpDist = Vec2f.EuclidianDistance(position, this[i]);

                if (tmpDist > tmpRadius)
                    tmpRadius = tmpDist;

            }

            radius = tmpRadius;
        }


        public override ACollider Copy()
        {
            Vector2f[] vertices = new Vector2f[this.vertices.Count];

           for (int i = 0; i < vertices.Length; i++)
               vertices[i] = this[i];

           return new ConvexCollider(this.redundantNormals, vertices);
        }


        //TODO: test
        public override bool IntersectsPoint(Vector2f point)
        {
            for (int i = 0; i < VertexCount; i++)
            {
                if (Vec2f.Dot(point, normals[i]) - Vec2f.Dot(this[i], normals[i]) >= 0)
                    return false;
            }

            return true;
        }


        public override void DebugDraw(RenderTarget target, Color color)
        {

            //collision shape:

            VertexArray vertexArray = new VertexArray(PrimitiveType.TrianglesFan,(uint)(this.vertices.Count + 2));

            vertexArray[0] = new Vertex(this.position, color);

            int vertexCounter = 0;
            for (uint i = 1; i < vertexArray.VertexCount - 1; i++)
            {
                vertexArray[i] = new Vertex(this.vertices[vertexCounter++], color);
            }

            vertexArray[vertexArray.VertexCount - 1] = vertexArray[1];
            target.Draw(vertexArray);


            //axis aligned bounding rectangle:
            /*
            RectangleShape boundingRectangle = new RectangleShape(new Vector2f(boundingRect.Width, boundingRect.Height));
            boundingRectangle.Position = position;
            boundingRectangle.Origin = new Vector2f(boundingRect.Width * 0.5f, boundingRect.Height * 0.5f);
            boundingRectangle.FillColor = new Color(0, 255, 0, 128);
            target.Draw(boundingRectangle);
            */
            //bounding Circle:
            /*
            CircleShape boundingCircle = new CircleShape(this.radius);
            boundingCircle.Position = position;
            boundingCircle.Origin = new Vector2f(radius, radius);
            boundingCircle.FillColor = new Color(0, 0, 255, 128);
            target.Draw(boundingCircle);
             */
        }
    }
}
