﻿using SFML.Graphics;
using SFML.Window;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace BobbleCrowd
{
    public struct Matrix
    {
        /// <summary>
        /// Matrix that does nothing on transformation -> [1 0 \n 0 1]
        /// </summary>
        public static Matrix Identity
        {
            get
            {
                return new Matrix(1.0f, 0.0f, 0.0f,
                                  0.0f, 1.0f, 0.0f);
            }
        }

        float[] a;

        private Matrix(float a, float b, float c, float d, float e, float f)
        {
            this.a = new float[6] { a, b, c, 
                                    d, e, f };
        }

        /// <summary>
        /// Converts the current Matrix2D to a rotation matrix with a rotation by rad radians.
        /// </summary>
        /// <param name="rad"></param>
        public void ConvertToRotation(float rad)
        {
            float sin = (float)Math.Sin(rad);
            float cos = (float)Math.Cos(rad);

            a[0] = cos;
            a[1] = -sin;
            a[2] = 0;
            a[3] = sin;
            a[4] = cos;
            a[5] = 0;
        }

        /// <summary>
        /// Converts the current Matrix2D to a scaling matrix with dX and dY scaling in percentage.
        /// </summary>
        /// <param name="dX"></param>
        /// <param name="dY"></param>
        public void ConvertToScale(float dX, float dY)
        {
            a[0] = dX;
            a[1] = 0;
            a[2] = 0;
            a[3] = 0;
            a[4] = dY;
            a[5] = 0;
        }

        public void ConverToTranslation(float dX, float dY)
        {
            a[0] = 1.0f;
            a[1] = 0.0f;
            a[2] = dX;
            a[3] = 0.0f;
            a[4] = 1.0f;
            a[5] = dY;
        }


        /// <summary>
        /// Multiplicates the given Matrix2D with a Vector2f (applying the transformation to the vector).
        /// </summary>
        /// <param name="v">The vector to be transformed.</param>
        /// <returns>The transformed vector.</returns>
        public Vector2f Transform(Vector2f v)
        {
            return new Vector2f(a[0] * v.X + a[1] * v.Y + a[2], 
                                a[3] * v.X + a[4] * v.Y + a[5]);
        }

        public Matrix Mult(Matrix b)
        {
            return new Matrix(  a[0] * b.a[0] + a[1] * b.a[3],    a[0] * b.a[1] + a[1] * b.a[4],        a[0] * b.a[2] + a[1] * b.a[5] + a[2],
                                a[2] * b.a[0] + a[4] * b.a[3],    a[3] * b.a[1] + a[4] * b.a[4],        a[3] * b.a[2] + a[4] * b.a[5] + a[5]);
        }

        public Matrix Add(Matrix b)
        {
            return new Matrix(  a[0] + b.a[0], a[1] + b.a[1], a[2] + b.a[2],
                                a[3] + b.a[3], a[4] + b.a[4], a[5] + b.a[5]);
        }

        public Matrix MultComp(Matrix b)
        {
            return new Matrix(  a[0] * b.a[0], a[1] * b.a[1], a[2] * b.a[2],
                                a[3] * b.a[3], a[4] * b.a[4], a[5] * b.a[5]);
        }

        public Matrix Sub(Matrix b)
        {
            return new Matrix(  a[0] - b.a[0], a[1] - b.a[1], a[2] - b.a[2],
                                a[3] - b.a[3], a[4] - b.a[4], a[5] - b.a[5]);
        }


        /// <summary>
        /// Creates a Matrix2D that can scale vectors.
        /// </summary>
        /// <param name="dX"></param>
        /// <param name="dY"></param>
        /// <returns></returns>
        public static Matrix CreateScale(float dX, float dY)
        {
            return new Matrix(dX,   0.0f,   0.0f,
                              0.0f, dY,     0.0f);
        }

        /// <summary>
        /// Creates a Matrix2D that can rotate vectors.
        /// </summary>
        /// <param name="rad"></param>
        /// <returns></returns>
        public static Matrix CreateRotation(float rad)
        {
            float sin = (float)Math.Sin(rad);
            float cos = (float)Math.Cos(rad);

            return new Matrix(  cos,    -sin,   0.0f,
                                sin,    cos,    0.0f);
        }

        public static Matrix CreateTranslation(float dx, float dy){
            return new Matrix(  1.0f, 0.0f, dx,
                                0.0f, 1.0f, dy);
        }


    }
}
