﻿
using SFML.Graphics;
using SFML.Window;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml;

namespace BobbleCrowd.Tiled
{
    public class LevelReader
    {
        public int numTilesX = 0;
        public int numTilesY = 0;

        public int tileWidth = 0;
        public int tileHeight = 0;

        public int chunkWidth = 0;
        public int chunkHeight = 0;


        public TileSetManager tileSetManager = new TileSetManager();

        public List<List<TileData>> tileData = new List<List<TileData>>();

        public List<TileData> decorationData = new List<TileData>();

        public List<List<Vector2f>> collisions = new List<List<Vector2f>>();
        public List<String> collisionName = new List<String>();


        private int tileLayersCounter = 0;

        public List<FloatRect> redZoneRect = new List<FloatRect>();
        public List<FloatRect> blueZoneRect = new List<FloatRect>();
        public List<FloatRect> greenZoneRect = new List<FloatRect>();



        public LevelReader()
        {

        }


        public void Load(String name)
        {
            XmlReader reader = XmlReader.Create(name);

            while (reader.Read())
            {
                if (reader.Name.Equals("map"))
                    break;
            }

            LoadMap(reader);
        }

        private void LoadMap(XmlReader reader)
        {
            //print(reader.Name);
            while (reader.MoveToNextAttribute())
            {
                if (reader.Name.Equals("width"))
                    numTilesX = int.Parse(reader.Value);
                else if (reader.Name.Equals("height"))
                    numTilesY = int.Parse(reader.Value);
                else if (reader.Name.Equals("tilewidth"))
                    tileWidth = int.Parse(reader.Value);
                else if (reader.Name.Equals("tileheight"))
                    tileHeight = int.Parse(reader.Value);
            }

            while (reader.Read())
            {
                if (reader.NodeType != XmlNodeType.EndElement)
                {
                    if (reader.Name.Equals("properties"))
                        LoadMapProps(reader);

                    else if (reader.Name.Equals("tileset"))
                        LoadTileSet(reader);

                    else if (reader.Name.Equals("layer"))
                        LoadLayer(reader);

                    else if (reader.Name.Equals("objectgroup"))
                        LoadObjectGroup(reader);
         
                }

            }
        }

        private void LoadObjectGroup(XmlReader reader)
        {
            while (reader.MoveToNextAttribute())
            {
                //Console.WriteLine(reader.Name + " = " + reader.Value);

                if (reader.Value.StartsWith("Decoration"))
                    LoadDecorationLayer(reader);

                else if (reader.Value.StartsWith("Region"))
                    LoadRegionLayer(reader);

                else if (reader.Value.StartsWith("Collision"))
                    LoadCollisionLayer(reader);
            }
        }

        private void LoadRegionLayer(XmlReader reader)
        {
            while (reader.Read() && !(reader.NodeType == XmlNodeType.EndElement && reader.Name.Equals("objectgroup")))
            {
                if (reader.Name.Equals("object"))
                {
                    FloatRect rect = new FloatRect();
                    BodyColor color = BodyColor.Count;

                    while (reader.MoveToNextAttribute())
                    {
                        if (reader.Name.Equals("type"))
                        {
                            String type = reader.Value;

                            if (type.Equals("red"))
                                color = BodyColor.Red;

                            else if (type.Equals("blue"))
                                color = BodyColor.Blue;

                            else if (type.Equals("green"))
                                color = BodyColor.Green;
                        }

                        else if (reader.Name.Equals("x"))
                            rect.Left = float.Parse(reader.Value, Settings.floatParseFormat);

                        else if (reader.Name.Equals("y"))
                            rect.Top = float.Parse(reader.Value, Settings.floatParseFormat);

                        else if (reader.Name.Equals("width"))
                            rect.Width = float.Parse(reader.Value, Settings.floatParseFormat);

                        else if (reader.Name.Equals("height"))
                            rect.Height = float.Parse(reader.Value, Settings.floatParseFormat);
                    }

                    Debug.Assert(color != BodyColor.Count);

                    if (color == BodyColor.Blue)
                        blueZoneRect.Add(rect);

                    else if (color == BodyColor.Green)
                        greenZoneRect.Add(rect);

                    else if (color == BodyColor.Red)
                        redZoneRect.Add(rect);
                }


            }
        }

        private void LoadCollisionLayer(XmlReader reader)
        {
            char[] splitSymbolsSpace = {' '};
            char[] splitSymbolsComma = {','};
            
            while (reader.Read() && !(reader.NodeType == XmlNodeType.EndElement && reader.Name.Equals("objectgroup")))
            {

                if (reader.Name.Equals("object"))
                {
                    Vector2f start = new Vector2f();
                    float rotation = 0.0f;
                    String name = null;
                    List<Vector2f> polygon = new List<Vector2f>();

                    while (reader.MoveToNextAttribute())
                    {
                        if (reader.Name.Equals("type"))
                            name = reader.Value;

                        else if (reader.Name.Equals("x"))
                            start.X = float.Parse(reader.Value, Settings.floatParseFormat);

                        else if (reader.Name.Equals("y"))
                            start.Y = float.Parse(reader.Value, Settings.floatParseFormat);

                        else if (reader.Name.Equals("rotation"))
                            rotation = float.Parse(reader.Value, Settings.floatParseFormat);

                    }

                    collisionName.Add(name);

                    while (reader.Read() && !(reader.NodeType == XmlNodeType.EndElement && reader.Name.Equals("object")))
                    {
                        while (reader.MoveToNextAttribute())
                        {

                            if (reader.Name.Equals("points"))
                            {
                                String[] vectors = reader.Value.Split(splitSymbolsSpace, StringSplitOptions.RemoveEmptyEntries);

                                for (int i = 0; i < vectors.Length; i++)
                                {
                                    String[] coords = vectors[i].Split(splitSymbolsComma, StringSplitOptions.RemoveEmptyEntries);

                                    polygon.Add(new Vector2f(   float.Parse(coords[0], Settings.floatParseFormat), 
                                                                float.Parse(coords[1],Settings.floatParseFormat)) + start);
                                }
                            }

                        }
                    }


                    //TODO: watch this part: is it always the right origin to rotate from?
                    if (rotation != 0.0f)
                        Vec2f.RotateAllFromOrigin(polygon, MathUtil.ToRad(rotation), polygon[0]);

                    collisions.Add(polygon);


                }


            }
        }

        private void LoadDecorationLayer(XmlReader reader)
        {
      
            while (reader.Read() && !(reader.NodeType == XmlNodeType.EndElement && reader.Name.Equals("objectgroup")))
            {
                if (reader.Name.Equals("object"))
                {
                    Vector2f pos = new Vector2f();
                    int id = -1;
                    float rotation = 0.0f;

                    while (reader.MoveToNextAttribute())
                    {
                        if (reader.Name.Equals("gid"))
                            id = int.Parse(reader.Value);

                        else if (reader.Name.Equals("x"))
                            pos.X = float.Parse(reader.Value, Settings.floatParseFormat);

                        else if (reader.Name.Equals("y"))
                            pos.Y = float.Parse(reader.Value, Settings.floatParseFormat);

                        else if (reader.Name.Equals("rotation"))
                            rotation = float.Parse(reader.Value, Settings.floatParseFormat);

                    }

                    decorationData.Add(new TileData(pos, id, rotation));
                }

            }
        }

        private void LoadLayer(XmlReader reader)
        {
            tileData.Add(new List<TileData>());

            int tileCount = 0;

            while (reader.Read() && !(reader.NodeType == XmlNodeType.EndElement && reader.Name.Equals("layer")))
            {
                 while (reader.MoveToNextAttribute())
                {
                    if (reader.Name.Equals("gid"))
                    {
                        int id = int.Parse(reader.Value);

                        int worldX = (tileCount % numTilesX) * tileWidth;
                        int worldY = (tileCount / numTilesX) * tileHeight;

                        TileData tmpData = new TileData(worldX, worldY, id, 0.0f);

                        tileData[tileLayersCounter].Add(tmpData);

                        ++tileCount;
                    }
                 }
            }

            tileLayersCounter++;
        }

        private void LoadTileSet(XmlReader reader)
        {
            String tileSetName = null;
            int firstId = -1;

            int currentTileWidth = 0;
            int currentTileHeight = 0;

            while (reader.MoveToNextAttribute())
            { 
                if (reader.Name.Equals("firstgid"))
                    firstId = int.Parse(reader.Value);

                else if (reader.Name.Equals("tilewidth"))
                    currentTileWidth = int.Parse(reader.Value);

                else if (reader.Name.Equals("tileheight"))
                    currentTileHeight = int.Parse(reader.Value);
            }

            while (reader.Read() && !(reader.NodeType == XmlNodeType.EndElement && reader.Name.Equals("tileset")))
            {
                while (reader.MoveToNextAttribute())
                {
                    if (reader.Name.Equals("source"))
                        tileSetName = reader.Value;
                }
            }

            tileSetManager.AddTileSet(new TileSet(tileSetName, currentTileWidth, currentTileHeight, firstId));
        }

        private void LoadMapProps(XmlReader reader)
        {
            while (reader.Read() && !(reader.NodeType == XmlNodeType.EndElement && reader.Name.Equals("properties")))
            {
                while (reader.MoveToNextAttribute())
                {
                    if (reader.Value.Equals("chunkWidth"))
                    {
                        reader.MoveToNextAttribute();
                        chunkWidth = int.Parse(reader.Value);
                    }
                    else if (reader.Value.Equals("chunkHeight"))
                    {
                        reader.MoveToNextAttribute();
                        chunkHeight = int.Parse(reader.Value);
                    }
                }
            }
        }
        public void printAll()
        {
            Console.WriteLine("TileSizeX = " + tileWidth);
            Console.WriteLine("TileSizeY = " + tileHeight);

            Console.WriteLine("NumTilesX = " + numTilesX);
            Console.WriteLine("NumTilesY = " + numTilesY);

            Console.WriteLine("NumLayers = " + tileLayersCounter);


            Console.WriteLine("ChunkWidth = " + chunkWidth);
            Console.WriteLine("ChunkHeight = " + chunkHeight);

        }

    }
}


