﻿using SFML.Graphics;
using SFML.Window;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace BobbleCrowd.Tiled
{
    /// <summary>
    /// Class that handles the drawing of the background, static tiles. Will only draw tiles that are inside the view.
    /// </summary>
    public class TileMap
    {
        public TileSetManager tileSetManager;

        private Texture tileTexture;
        private Sprite tileSprite;

        /// <summary>
        /// Does nothing.
        /// </summary>
        public TileMap()
        {

        }

        /// <summary>
        /// Creates the tileMap for the given parameter.
        /// </summary>
        /// <param name="worldWidth">The total width of the world.</param>
        /// <param name="worldHeight">The total height of the world.</param>
        /// <param name="tileData">All layers of all tiles</param>
        /// <param name="tileSet">Defines what tiles to be drawn.</param>
        public void Create(int worldWidth, int worldHeight, List<List<Tiled.TileData>> tileData, TileSetManager tileSetManager, List<TileData> decoration)
        {
            //the place where all tiles will be rendered to
            RenderTexture tileTarget = new RenderTexture((uint)worldWidth, (uint)worldHeight);

            this.tileSetManager = tileSetManager;
            
            //foreach layer
            for (int i = 0; i < tileData.Count; i++)
            {
                //foreach tile in this layer
                for (int j = 0; j < tileData[i].Count; j++)
                {
                    Tiled.TileData tmpData = tileData[i][j];

                    //helper sprite to draw that tile to the tileTarget
                    Sprite sprite = new Sprite(tileSetManager.GetTexture(tmpData.id));
                    sprite.Position = tmpData.worldPos;
                    sprite.TextureRect = tileSetManager.GetSourceRectangle(tmpData.id);

                    tileTarget.Draw(sprite);
                }
            }

            //foreach decoration:
            for (int i = 0; i < decoration.Count; i++)
            {
                Sprite sprite = new Sprite(tileSetManager.GetTexture(decoration[i].id));

                sprite.TextureRect = tileSetManager.GetSourceRectangle(decoration[i].id);

                //since tiled uses lowerLeftPoint as origin..... >.<
                sprite.Origin = new Vector2f(0, tileSetManager.GetTileHeight(decoration[i].id));

                sprite.Position = decoration[i].worldPos;
                sprite.Rotation = decoration[i].rotation;

                tileTarget.Draw(sprite);
            }

            //finish up the tileTarget
            tileTarget.Display();

            //retrieve the texture from the target and copy it
            this.tileTexture = new Texture(tileTarget.Texture);

            //set up the sprite to be drawn
            this.tileSprite = new Sprite(this.tileTexture);

            //delete the tileTarget
            tileTarget.Dispose();
            tileTarget = null;

            tileSetManager.Dispose();
        }

        /// <summary>
        /// Draws all tiles that are within the view of the given RenderTarget.
        /// </summary>
        /// <param name="target">Where to draw the tiles to. Also uses its View to compute which tiles to be drawn.</param>
        public void Draw(RenderTarget target)
        {
            View v = target.GetView();

            //only needed when the view can rotate:
            float tmp = 1.75f;

            Vector2f upperLeftCorner = v.Center - v.Size * tmp * 0.5f;
            Vector2f size = v.Size * tmp;

            this.tileSprite.Position = upperLeftCorner;
            this.tileSprite.TextureRect = new IntRect((int)(upperLeftCorner.X), (int)(upperLeftCorner.Y), (int)(size.X), (int)(size.Y) );

            target.Draw(this.tileSprite);
        }


    }
}
