﻿using BobbleCrowd.Components;
using BobbleCrowd.World;
using SFML.Graphics;
using SFML.Window;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace BobbleCrowd.Entities
{
    //TODO: maybe other names, more flags:
    [Flags]
    public enum CollisionGroup
    {
        None            = 0,

        Group1          = 1,
        Group1Bullet    = 2,

        Group2          = 4,
        Group2Bullet    = 8,
        
        Group3          = 16,
        Group3Bullet    = 32,

        Environment     = 64,
    };



    public class Entity : IUnique, INameable
    {
        #region Member

        public String Name
        {
            get;
            set;
        }

        /// <summary> If its true, this Entity will be removed at the end of this Update cycle.</summary>
        public bool IsDead
        {
            get;
            set;
        }

        /// <summary> Specifies to which group this Entity belongs to. </summary>
        public CollisionGroup Category
        {
            get;
            set;
        }

        public bool updateAtAll = true;

        /// <summary> Specifies with which groups this Entity can collide.</summary>
        public CollisionGroup Mask
        {
            get;
            set;
        }

        /// <summary> UniqueId for the whole GameWorld </summary>
        public int UniqueId { get; set; }

        /// <summary> The indices in which Chunk this entity is.</summary>
        public Vector2i ChunkId { get; set; }

        /// <summary> The type of ACollider this entity owns and that is used for collision detection. </summary>
        public ACollider Collider { get; private set; }

        /// <summary> If this entity is about to leave its current chunk.</summary>
        public bool LeavesChunk { get; set; }


        //______________THINKER____________
        private AThinker thinker;
        public AThinker Thinker
        {
            get { return thinker; }
            set { SetThinker(value); }
           
        }
        private void SetThinker(AThinker value)
        {
            if (value == null)
                thinker = AThinker.GetEmpty();

            else
            {
                value.Owner = this;
                value.IsActive = true;
                thinker = value;
            }

            thinker.Init();
        }


        //______________RESPONSE____________
        private ACollisionResponse response;
        public ACollisionResponse Response
        {
            get { return response; }
            set { SetResponse(value); }

        }
        private void SetResponse(ACollisionResponse value)
        {
            if (value == null)
                response = ACollisionResponse.GetEmpty();

            else
            {
                value.Owner = this; 
                value.IsActive = true; 
                response = value;
            }

            response.Init();
         
        }


        //______________DRAWER____________
        private ADrawer drawer;
        public ADrawer Drawer
        {
            get { return drawer; }
            set { SetDrawer(value); }
        }
        private void SetDrawer(ADrawer value)
        {
            if (value == null)
                drawer = ADrawer.GetEmpty(true);

            else
            {
                value.Owner = this;
                value.IsActive = true;
                drawer = value;
            }

            drawer.Init();
        }


        //


 
        public GameWorld World
        {
            get;
            set;
        }
        #endregion

        public bool CanCollide { get; set; }

        public Entity(String name, ACollider collider, Vector2f position)
        {
            this.Name = name;

            this.Collider = collider;

            this.Collider.Position = position;

            UniqueId = -1;

            ChunkId = new Vector2i(0, 0);

            LeavesChunk = false;

            InitEmptyComponents();

            this.CanCollide = true;
        }


        public void InitEmptyComponents()
        {
            Thinker = AThinker.GetEmpty();

            Response = ACollisionResponse.GetEmpty();

            Drawer = ADrawer.GetEmpty(true);
        }

        public virtual void Update()
        {
            if(thinker.IsActive)
                thinker.Update();

            if(response.IsActive)
                response.Update();

            if (drawer.IsActive)
                drawer.Update();
            
        }

        /// <summary>
        /// Tests if this Entity left its current chunk and handles the leaving aswell.
        /// </summary>
        /// <param name="world"></param>
        public void TestForLeavingChunk()
        {
            //the new chunk:
            Vector2i maybeNewChunkId = new Vector2i((int)(Collider.Position.X / this.World.ChunkWidth), (int)(Collider.Position.Y / this.World.ChunkHeight));

            if (!MathUtil.IsEqual(maybeNewChunkId, ChunkId))
            {
                //notify that the entity on the old chunk, wants to leave:
                this.World.NotifyLeaving(this.ChunkId, this);

                this.ChunkId = maybeNewChunkId;

                this.LeavesChunk = true;
            }
        }

        public void Draw(GameTime gameTime, RenderTarget target)
        {
            if(drawer.toBeDrawn)
                drawer.Draw(target);
        }

        //TODO: call onDeath or kill it, no matter, what it is?
        public void Kill()
        {
            World.RemoveEntity(this);
        }










    }
}
