﻿using BobbleCrowd.Components;
using BobbleCrowd.Entities;
using BobbleCrowd.GameStates;
using BobbleCrowd.Tiled;
using BobbleCrowd.World;
using SFML.Graphics;
using SFML.Window;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using BobbleCrowd.Message;
using SFML.Audio;

namespace BobbleCrowd.World
{
    public sealed class GameWorld : TextureDrawGroup
    {
        public FloatRect gameArea;
        public GameMode mode;

        public bool endGame = false;

        public int deleteMode = -1;
        public float deleteTime = 0;
        public bool secondDel = false;

        public string LevelName
        {
            get;
            private set;
        }

        public GameTime Time
        {

            get;
            set;
        }

        #region WorldAttribs

        public TileMap tileMap = new TileMap();

        public int WorldWidth { get; private set; }
        public int WorldHeight { get; private set; }

        public int NumTilesX { get; private set; }
        public int NumTilesY { get; private set; }
        #endregion

        #region Entities

        /// <summary> A Queue that will be iterated after every update tick to check for intersections.</summary>
        private Queue<EntityChunkLookup> collidableEntities = new Queue<EntityChunkLookup>();

        private Queue<Entity> removingEntities = new Queue<Entity>();

        private UniqueList<Entity> entities = new UniqueList<Entity>();

        public void AddEntity(Entity e)
        {
            Debug.Assert(e.UniqueId == -1);

            e.World = this;

            entities.AddLast(e);

            Vector2i id = new Vector2i((int)(e.Collider.Position.X / ChunkWidth),
                                        (int)(e.Collider.Position.Y / ChunkHeight));

            e.ChunkId = id;

            chunks[id.X, id.Y].AddEntity(e);
        }

        private void FinalRemoveEntity(Entity e)
        {
            Debug.Assert(e.UniqueId != -1);

            entities.Remove(e);
            chunks[e.ChunkId.X, e.ChunkId.Y].RemoveEntity(e);
        }

        public void NotifyPossibleCollision(EntityChunkLookup lookup)
        {
            collidableEntities.Enqueue(lookup);
        }

        public void RemoveEntity(Entity e)
        {
            if (!e.IsDead)
            {
                removingEntities.Enqueue(e);
                e.IsDead = true;
            }

        }

        public Entity GetEntity(int uniqueId)
        {
            return entities.GetByUniqueId(uniqueId);
        }


        #endregion

        #region ChunkStuff

        /// <summary> The subdivion of the world in chunks. </summary>
        private Chunk[,] chunks;

        public int ChunkWidth { get; private set; }
        public int ChunkHeight { get; private set; }

        public int NumChunksX { get; private set; }
        public int NumChunksY { get; private set; }

        public Chunk GetChunkAt(float worldX, float worldY)
        {
            return chunks[(int)(worldX / ChunkWidth), (int)(worldY / ChunkHeight)];
        }
        public Chunk GetChunkAt(Vector2f worldPos)
        {
            return GetChunkAt(worldPos.X, worldPos.Y);
        }
        public Chunk GetChunkAt(int idX, int idY)
        {
            return chunks[idX, idY];
        }
        public Chunk GetChunkAt(Vector2i index)
        {
            return chunks[index.X, index.Y];
        }

        public bool IsValidChunk(Vector2i id)
        {
            return IsValidChunk(id.X, id.Y);
        }
        public bool IsValidChunk(int x, int y)
        {
            return IsValidChunkIdX(x) && IsValidChunkIdY(y);
        }
        public bool IsValidChunkIdY(int y)
        {
            return y >= 0 && y < NumChunksY;
        }
        public bool IsValidChunkIdX(int x)
        {
            return x >= 0 && x < NumChunksX;
        }


        /// <summary> All chunks that have currently at least one unit that leaves its Chunk. </summary>
        private Queue<EntityChunkLookup> chunkLeavingQueue = new Queue<EntityChunkLookup>();

        /// <summary>
        /// Notifies the GameWorld that there is a leaving Entity at the given chunk.
        /// </summary>
        /// <param name="chunkId">Where the Entity e was before leaving its current Chunk.</param>
        /// <param name="e">Which Entity left</param>
        public void NotifyLeaving(Vector2i chunkId, Entity e)
        {
            chunkLeavingQueue.Enqueue(new EntityChunkLookup(e.UniqueId, chunkId));
        }

        #endregion

        private InGame inGame;

        public InGame GetInGame()
        {
            return inGame;
        }

        public GameWorld()
            : base(Settings.WindowSize.X, Settings.WindowSize.Y, Strings.TEXTURE_DRAWER_GAMEWORLD)
        {

        }

        public void Create(GameTime gameTime, InGame inGame, ModeType type)
        {
            endGame = false;
            String mapName;
            if (type.Equals(ModeType.BOBBLE_MIX))
            {
                mode = new BobbleMixMode(this);
                mapName = "level.tmx";
            }
            else if (type.Equals(ModeType.BOBBLE_SORT))
            {
                mode = new BobbleSortMode(this);
                mapName = "levelSort.tmx";
            }
            else
            {
                throw new Exception("mhhh Bobble!  - Wrong Mode!");
            }

            this.inGame = inGame;

            this.Time = gameTime;

            this.LevelName = mapName;

            Tiled.LevelReader reader = new Tiled.LevelReader();
            reader.Load("Content/" + mapName);
            //  reader.printAll();

            //Compute how big the world needs to be:
            WorldWidth = reader.tileWidth * reader.numTilesX;
            WorldHeight = reader.tileHeight * reader.numTilesY;

            //______CHUNKS_____________
            ChunkWidth = reader.chunkWidth;
            ChunkHeight = reader.chunkHeight;

            NumChunksX = WorldWidth / reader.chunkWidth;
            NumChunksY = WorldHeight / reader.chunkHeight;

            chunks = new Chunk[NumChunksX, NumChunksY];

            //create empty chunks:
            for (int i = 0; i < chunks.GetLength(0); i++)
                for (int j = 0; j < chunks.GetLength(1); j++)
                    chunks[i, j] = new Chunk(i, j, this);

            //______TILES_____________
            NumTilesX = reader.numTilesX;
            NumTilesY = reader.numTilesY;

            this.tileMap.Create(WorldWidth, WorldHeight, reader.tileData, reader.tileSetManager, reader.decorationData);

            //create blocker things:
            for (int i = 0; i < reader.collisions.Count; i++)
            {
                Vector2f[] currentPoly = new Vector2f[reader.collisions[i].Count];
                Vector2f center = new Vector2f();

                for (int j = 0; j < reader.collisions[i].Count; j++)
                {
                    currentPoly[j] = reader.collisions[i][j];
                    center += currentPoly[j];
                }

                center /= currentPoly.Length;

                AddEntity(CreateInviBlocker(new ConvexCollider(currentPoly), center, reader.collisionName[i]));
            }


            mode.Init(reader);



        }

        private Entity CreateInviBlocker(ConvexCollider convexCollider, Vector2f center, string p)
        {
            Entity e = new Entity(p, convexCollider, center);

            e.Category = CollisionGroup.Environment;
            e.Mask = CollisionGroup.Group2;

            e.Response = new PushResponse(false, 1.0f);


            e.Drawer.toBeDrawn = false;

            return e;
        }


        public override void Update(GameTime gameTime)
        {

            if (!endGame)
            {


                mode.Update(gameTime);

                if(deleteMode==0)
                {
                    if (!secondDel)
                    {
                        if (deleteTime == 0)
                        {
                            for (int i = 0; i < entities.Count; i++)
                            {
                                if (entities[i] is Bobble && (entities[i] as Bobble).Category == Entities.CollisionGroup.Group3)
                                {

                                    RemoveBobble_SortMode(entities[i] as Bobble);
                                }
                            }
                        }
                        deleteTime += (float)gameTime.ElapsedTime.TotalMilliseconds;
                        if (deleteTime > 500)
                        {
                            secondDel = true;
                            deleteMode = 1;
                        }
                        
                    }
                    if (secondDel)
                    {
                        for (int i = 0; i < entities.Count; i++)
                        {
                            if (entities[i] is Bobble && (entities[i] as Bobble).Category == Entities.CollisionGroup.Group3)
                            {

                                RemoveBobble_SortMode(entities[i] as Bobble);
                            }
                        }
                        secondDel = false;
                        deleteMode = -1;
                        deleteTime = 0;
                    }
                }
                if(deleteMode ==1)
                {

                    if(secondDel)
                    {
                        for (int i = 0; i < entities.Count; i++)
                        {
                            if (entities[i] is Bobble && (entities[i] as Bobble).Category != Entities.CollisionGroup.Group3)
                            {

                                RemoveBobble_SortMode(entities[i] as Bobble);
                            }
                        }

                        secondDel = false;
                        deleteMode = -1;
                        deleteTime = 0;
                    }

                    if (!secondDel)
                    {
                        if (deleteTime == 0)
                        {
                            for (int i = 0; i < entities.Count; i++)
                            {
                                if (entities[i] is Bobble && (entities[i] as Bobble).Category != Entities.CollisionGroup.Group3)
                                {

                                    RemoveBobble_SortMode(entities[i] as Bobble);
                                }
                            }
                        }
                        deleteTime += (float)gameTime.ElapsedTime.TotalMilliseconds;
                        if(deleteTime>500)
                        {
                            secondDel = true;
                            deleteMode = 0;
                        }
                    }
                }
                
                foreach (Chunk c in chunks)
                {
                    c.Update();
                }



                //Test for colliding Entities:
                while (collidableEntities.Count > 0)
                {
                    EntityChunkLookup tmpLookup = collidableEntities.Dequeue();

                    Entity tmpEntity = entities.GetByUniqueId(tmpLookup.entityId);

                    if (tmpEntity.CanCollide)
                    {
                        for (int i = 0; i < chunks[tmpLookup.chunkId.X, tmpLookup.chunkId.Y].chunkEntities.Count; i++)
                        {

                            Entity tmpOtherEntity = chunks[tmpLookup.chunkId.X, tmpLookup.chunkId.Y].chunkEntities[i];

                            if (tmpOtherEntity == tmpEntity || !tmpOtherEntity.CanCollide || !tmpEntity.CanCollide)
                                continue;

                            else
                            {
                                //if the first entity can collide with the second entity
                                bool firstResult = (tmpEntity.Mask & tmpOtherEntity.Category) != CollisionGroup.None;

                                //if the second entity can colide with the first entity
                                bool secondResult = (tmpOtherEntity.Mask & tmpEntity.Category) != CollisionGroup.None;

                                //only test for intersection if at least one of the two entities can intersect at all
                                if (firstResult || secondResult)
                                {
                                    IntersectionData tmpData = tmpEntity.Collider.Intersects(tmpOtherEntity.Collider);

                                    if (tmpData.Intersects)
                                    {
                                        if (firstResult)
                                            tmpEntity.Response.OnCollision(tmpOtherEntity, tmpData, true);

                                        if (secondResult)
                                            tmpOtherEntity.Response.OnCollision(tmpEntity, tmpData, false);
                                    }
                                }
                            }
                        }

                    }

                }

                //handle leaving Entities:
                //as long as there are chunks that have leaving entities:
                while (chunkLeavingQueue.Count > 0)
                {
                    EntityChunkLookup currentLookUp = chunkLeavingQueue.Dequeue();

                    Entity e = entities.GetByUniqueId(currentLookUp.entityId);

                    if (e.LeavesChunk)
                    {
                        e.LeavesChunk = false;

                        if (e.Thinker is MouseClicker)
                        {
                            //WTF ?!!?!?!
                        }

                        //entity is inside the world:
                        else if (IsValidChunk(e.ChunkId) && e.Collider.Position.X > 0 && e.Collider.Position.Y > 0)
                        {
                            chunks[currentLookUp.chunkId.X, currentLookUp.chunkId.Y].chunkEntities.Remove(e);
                            chunks[e.ChunkId.X, e.ChunkId.Y].AddEntity(e);
                        }

                        //entity has left the world, just delete it, to avoid exceptions:
                        else
                        {
                            e.ChunkId = currentLookUp.chunkId;

                            if (e is Bobble)
                            {
                                if (mode is BobbleMixMode)
                                    (mode as BobbleMixMode).bobbleThinker.RemoveInfo((e as Bobble).info);
                            }

                            RemoveEntity(e);
                        }
                    }

                }

                //As long as there are entities, that need to be removed:
                while (removingEntities.Count > 0)
                    FinalRemoveEntity(removingEntities.Dequeue());
            }


        }
        public void RemoveAllBobble(bool lifeTimeDead)
        {
            if (mode is BobbleSortMode)
            {
                if (lifeTimeDead)
                    deleteMode = 0;
                else
                    deleteMode = 1;


            }
            else
            {
                throw new NotImplementedException("This functions works online for BobbleSortMode.");
            }

        }
        private void RemoveBobble_SortMode(Bobble bobble)
        {
            createBobbleParticle_SortMode(bobble.Collider.Position, bobble.info);
            bobble.Kill();
            //bobbleThinker.RemoveInfo(bobble.info);

            //bobbleThinker.TestChange();
        }
        private void createBobbleParticle_SortMode(Vector2f position, BobbleInfo info)
        {

            for (int i = 0; i < 15; i++)
            {
                float scale = 0.3f;
                float size = scale * 48.0f;

                Entity particle = new Entity("Particle", new CircleCollider(size), position);


                MultiThinker thinker = new MultiThinker();
                particle.Thinker = thinker;


                thinker.AddThinker(new LinearSlowDownMover(position, Vec2f.RandomDir, 2 + MathUtil.Rand.NextFloat() * 2));

                particle.Category = CollisionGroup.None;
                particle.Mask = CollisionGroup.None;
                particle.CanCollide = false;

                //particle.Response = new PushResponse(true, 0.5f);

                BobbleInfo cInfo = new BobbleInfo();

                cInfo.color = info.color;
                cInfo.eye = EyeType.Count;
                cInfo.mouth = MouthType.Count;

                particle.Drawer = new BobbleDraw(scale, cInfo);
                //bobble.Drawer = new BasicTextureDrawer(Game.Content.Load<Texture>("gfx/bodySprite.png"), false, true, scale, new IntRect(0, 0, 128, 128));

                AddEntity(particle);
            }

        }
        public override void Draw(GameTime gameTime)
        {
            target.Clear();

            tileMap.Draw(target);

            mode.PreChunkDraw(target);

            foreach (Chunk c in chunks)
                c.Draw(gameTime, target);

            mode.PostChunkDraw(target);

            

         //   DrawBoundingsOfChunks(gameTime, target);
            

            target.Display();
        }




        /// <summary>
        /// Draws the boundings of a Chunks, use for debug purposes only.
        /// </summary>
        /// <param name="target"></param>
        private void DrawBoundingsOfChunks(GameTime gameTime, RenderTarget target)
        {
            foreach (Chunk c in chunks)
            {
                c.DrawBoundings(target);
            }

        }

        public Entity CreateZone(FloatRect rect, BodyColor color)
        {
            Vector2f pos = new Vector2f(rect.Left + rect.Width * 0.5f, rect.Top + rect.Height * 0.5f);

            Entity zone = new Entity("Zone:" + color, new ConvexCollider(true, PolygonFactory.GetBox(rect.Width * 0.5f, rect.Height * 0.5f)), pos);
            
            zone.Response = new SortZoneResponse(color);

            zone.Category = CollisionGroup.Environment;

            zone.Mask = CollisionGroup.Group2;

            zone.Drawer.toBeDrawn = false;

            return zone;
        }



        /// <summary>Is called after InGame has loaded everything, and after GameWorld.Create()...</summary>
        public void Init()
        {

        }




        
        
        









    }
    public enum ModeType
    {
        BOBBLE_MIX,
        BOBBLE_SORT,
        count
    }
}
